# AUTOGENERATED! DO NOT EDIT! File to edit: nbs/04-client-prep.ipynb (unless otherwise specified).

__all__ = ['test_endpoints', 'construct_method_to_params_dict', 'construct_request_type_filter',
           'check_request_type_filter', 'determine_request_type_from_fields', 'determine_method_request_types',
           'construct_method_to_params_map', 'construct_method_info_dict']

# Cell
from tqdm import tqdm
from warnings import warn
from functools import reduce

from . import rawgen, specgen, raw, utils

# Cell
def test_endpoints(
    default_kwargs: dict
):
    methods_to_test = [func for func in dir(raw) if 'get_' in func]
    stream_to_df = dict()

    for method_to_test in tqdm(methods_to_test):
        method_func = getattr(raw, method_to_test)
        func_kwargs = dict(zip(method_func.__code__.co_varnames, method_func.__defaults__))

        for kwarg, value in default_kwargs.items():
            if kwarg in func_kwargs.keys():
                func_kwargs.update({kwarg: value})

        r = method_func(**func_kwargs)
        df = utils.parse_xml_response(r)

        stream_to_df[method_to_test.split('_')[1]] = df

    streams_without_content = []

    for stream, df in stream_to_df.items():
        if df.size == 0:
            streams_without_content += [stream]

    return streams_without_content

    if len(streams_without_content) > 0:
        warn(f"The following data streams returned no content data: {', '.join(streams_without_content)}")

    return stream_to_df

# Cell
def construct_method_to_params_dict(API_yaml):
    method_to_params = reduce(lambda k, v: {**k, **v}, [
        {
            f'{k}_{rawgen.clean_path_name(stream)}': {
                parameter['name']: rawgen.extract_parameter_example(parameter)
                for parameter
                in v['parameters']
            }
            for k, v
            in method.items()
            if k in ['get', 'post']
        }
        for stream, method
        in API_yaml['paths'].items()
    ])

    return method_to_params

# Cell
def construct_request_type_filter(
    has_start_time: bool,
    has_end_time: bool,
    has_start_date: bool,
    has_end_date: bool,
    has_date: bool,
    has_SP: bool,
    has_year: bool,
    has_month: bool,
    has_week: bool
):
    request_type_filter = {
        'year': (has_year + has_month + has_week == 1) and (has_year == 1),
        'month': (has_year + has_month == 1) and (has_month == 1),
        'week': (has_year + has_week == 1) and (has_week == 1),
        'year_and_month': has_year + has_month == 2,
        'year_and_week': has_year + has_week == 2,
        'SP_and_date': has_SP + has_date == 2,
        'date_range': has_start_time + has_end_time + has_start_date + has_end_date == 2,
        'date_time_range': has_start_time + has_end_time + has_start_date + has_end_date == 4,
        'non_temporal': has_start_time + has_end_time + has_start_date + has_end_date + has_SP + has_date + has_year + has_month == 0,
    }

    return request_type_filter

def check_request_type_filter(
    field_names: list,
    request_type_filter: dict,
    has_start_time: bool,
    has_end_time: bool,
    has_start_date: bool,
    has_end_date: bool,
    has_date: bool,
    has_SP: bool,
    has_year: bool,
    has_month: bool,
    has_week: bool
):
    """
    Checks the validity of the specified stream parameters

    The following conditions will raise an error:
    * has month without a year
    * has only one of start/end time
    * has only one of start/end date
    * has only one settlement period or date
    * filter does not contain only one request type
    """

    filter_str = f'\n\nFilter:\n{request_type_filter}\n\nField Names:\n{", ".join(field_names)}'

    assert {(False, True): True, (False, False): False, (True, True): False, (True, False): False}[(has_year, has_month)] == False, 'Cannot provide a month without a year' + filter_str
    assert {(False, True): True, (False, False): False, (True, True): False, (True, False): False}[(has_year, has_week)] == False, 'Cannot provide a week without a year' + filter_str
    assert has_start_time + has_end_time != 1, 'Only one of start/end time was provided' + filter_str
    assert has_start_date + has_end_date != 1, 'Only one of start/end date was provided' + filter_str
    assert (has_SP + has_date != 1) or (has_start_date + has_end_date == 2), 'Only one of date/SP was provided' + filter_str
    assert sum(request_type_filter.values()) == 1, 'Request type could not be determined\n\nFilter' + filter_str

    return

def determine_request_type_from_fields(
    field_names: list,
    start_time_cols: list=['StartTime'],
    end_time_cols: list=['EndTime'],
    start_date_cols: list=['StartDate', 'FromSettlementDate', 'FromDate'],
    end_date_cols: list=['EndDate', 'ToSettlementDate', 'ToDate'],
    date_cols: list=['SettlementDate', 'ImplementationDate', 'DecommissioningDate', 'Date', 'startTimeOfHalfHrPeriod'],
    SP_cols: list=['SettlementPeriod', 'Period'],
    year_cols: list=['Year'],
    month_cols: list=['Month', 'MonthName'],
    week_cols: list=['Week']
):
    has_start_time = bool(set(field_names).intersection(set(start_time_cols)))
    has_end_time = bool(set(field_names).intersection(set(end_time_cols)))
    has_start_date = bool(set(field_names).intersection(set(start_date_cols)))
    has_end_date = bool(set(field_names).intersection(set(end_date_cols)))
    has_date = bool(set(field_names).intersection(set(date_cols)))
    has_SP = bool(set(field_names).intersection(set(SP_cols)))
    has_year = bool(set(field_names).intersection(set(year_cols)))
    has_month = bool(set(field_names).intersection(set(month_cols)))
    has_week = bool(set(field_names).intersection(set(week_cols)))

    request_type_filter = construct_request_type_filter(
        has_start_time, has_end_time, has_start_date,
        has_end_date, has_date, has_SP, has_year, has_month, has_week
    )

    check_request_type_filter(
        field_names, request_type_filter, has_start_time, has_end_time, has_start_date,
        has_end_date, has_date, has_SP, has_year, has_month, has_week
    )

    request_type = [k for k, v in request_type_filter.items() if v==True][0]

    return request_type

# Cell
def determine_method_request_types(method_to_params):
    method_to_request_type = dict()

    for method in method_to_params.keys():
        field_names = list(method_to_params[method].keys())
        method_to_request_type[method] = determine_request_type_from_fields(field_names)

    return method_to_request_type

# Cell
def construct_method_to_params_map(method_to_params):
    standardised_params_map = {
        'start_time': ['StartTime'],
        'end_time': ['EndTime'],
        'start_date': ['StartDate', 'FromSettlementDate', 'FromDate'],
        'end_date': ['EndDate', 'ToSettlementDate', 'ToDate'],
        'date': ['SettlementDate', 'ImplementationDate', 'DecommissioningDate', 'Date', 'startTimeOfHalfHrPeriod'],
        'SP': ['SettlementPeriod', 'Period'],
        'year': ['Year'],
        'month': ['Month', 'MonthName'],
        'week': ['Week']
    }

    method_to_params_map = dict()

    for method, params in method_to_params.items():
        method_to_params_map[method] = dict()

        for param in params.keys():
            for standardised_param, bmrs_params in standardised_params_map.items():
                if param in bmrs_params:
                    method_to_params_map[method][standardised_param] = param

    return method_to_params_map

# Cell
def construct_method_info_dict(API_yaml_fp: str):
    API_yaml = specgen.load_API_yaml(API_yaml_fp)

    method_to_params = construct_method_to_params_dict(API_yaml)
    method_to_request_type = determine_method_request_types(method_to_params)
    method_to_params_map = construct_method_to_params_map(method_to_params)

    method_info = dict()

    for method, params in  method_to_params.items():
        method_info[method] = dict()

        method_info[method]['request_type'] = method_to_request_type[method]
        method_info[method]['kwargs_map'] = method_to_params_map[method]
        method_info[method]['func_kwargs'] = {
            (
                {v: k for k, v in method_to_params_map[method].items()}[k]
                if k in method_to_params_map[method].values()
                else k
            ): v
            for k, v
            in method_to_params[method].items()
        }

    return method_info


# AUTOGENERATED! DO NOT EDIT! File to edit: nbs/09-map-gen.ipynb (unless otherwise specified).

__all__ = ['construct_df_PN_pivot_dt_rng', 'construct_PN_pivot_df', 'get_latest_PN_df', 'construct_osuked_id_mappings',
           'construct_map_df', 'df_to_gdf', 'generate_map_js', 'generate_map_md', 'app', 'generate_map']

# Cell
import json
import numpy as np
import pandas as pd
import geopandas as gpd

import os
import typer
from tqdm import tqdm
from jinja2 import Template

from . import utils, raw

# Cell
def construct_df_PN_pivot_dt_rng(df_PN):
    no_seconds = (((df_PN['timeFrom'].str.split(':').str[-1]=='00').mean()==1) &
                  ((df_PN['timeTo'].str.split(':').str[-1]=='00').mean()==1))

    if no_seconds == True:
        dt_rng = pd.date_range(df_PN['timeFrom'].min(), df_PN['timeTo'].max(), freq='min')
    else:
        dt_rng = pd.date_range(df_PN['timeFrom'].min(), df_PN['timeTo'].max(), freq='s')

    return dt_rng

def construct_PN_pivot_df(df_PN):
    bmu_ids = sorted(list(df_PN['bmUnitID'].unique()))
    df_PN_pivot = pd.DataFrame(index=construct_df_PN_pivot_dt_rng(df_PN), columns=bmu_ids, dtype=float)

    for bmu_id in tqdm(bmu_ids):
        for idx, row in df_PN.query('bmUnitID==@bmu_id').iterrows():
            df_PN_pivot.loc[pd.to_datetime(row['timeFrom']), bmu_id] = float(row['pnLevelFrom'])
            df_PN_pivot.loc[pd.to_datetime(row['timeTo']), bmu_id] = float(row['pnLevelTo'])

        df_PN_pivot[bmu_id] = df_PN_pivot[bmu_id].interpolate()

    return df_PN_pivot

def get_latest_PN_df(api_key):
    start_date = pd.Timestamp.now().round('30min') - pd.Timedelta(minutes=30)
    end_date = pd.Timestamp.now().round('30min') + pd.Timedelta(minutes=90)

    df = pd.DataFrame()

    for idx, (date, SP) in tqdm(utils.dt_rng_to_SPs(start_date, end_date).iterrows()):
        df_SP = utils.parse_xml_response(raw.get_PHYBMDATA(api_key, SettlementDate=date, SettlementPeriod=SP, ServiceType='xml'))
        df = df.append(df_SP)

    df = (df
          .query('recordType=="PN"')
          .dropna(how='all', axis=1)
          .pipe(construct_PN_pivot_df)
         )

    return df

# Cell
def construct_osuked_id_mappings(df_powerdict):
    osuked_id_mappings = dict()

    osuked_id_mappings['bmu_ids'] = (df_powerdict
                                     .set_index('osuked_id')
                                     ['sett_bmu_id']
                                     .str.split(', ')
                                     .dropna()
                                     .to_dict()
                                    )

    osuked_id_mappings['fuel_type'] = (df_powerdict
                                       .set_index('osuked_id')
                                       ['fuel_type']
                                       .dropna()
                                       .to_dict()
                                      )

    osuked_id_mappings['name'] = (df_powerdict
                                  .set_index('osuked_id')
                                  ['name']
                                  .dropna()
                                  .to_dict()
                                 )

    osuked_id_mappings['lat_lon'] = (df_powerdict
                                     .set_index('osuked_id')
                                     [['latitude', 'longitude']]
                                     .dropna()
                                     .apply(dict, axis=1)
                                     .to_dict()
                                    )

    return osuked_id_mappings

# Cell
def construct_map_df(
    s_PN,
    most_recent_available_dt,
    osuked_id_to_bmu_ids,
    osuked_id_to_lat_lon,
    osuked_id_to_fuel_type,
    osuked_id_to_name
):
    sites_data = list()

    for osuked_id, bmu_ids in osuked_id_to_bmu_ids.items():
        lat_lon_match = osuked_id in osuked_id_to_lat_lon.keys()

        matching_output_bmu_ids = s_PN.index.intersection(bmu_ids)
        output_match = matching_output_bmu_ids.size > 0

        if lat_lon_match and output_match:
            site_data = osuked_id_to_lat_lon[osuked_id]
            site_data.update({'osuked_id': osuked_id})
            site_data.update({'name': osuked_id_to_name[osuked_id]})
            site_data.update({'fuel_type': osuked_id_to_fuel_type[osuked_id]})
            site_data.update({'output': s_PN[matching_output_bmu_ids].sum()})
            site_data.update({'output_time': most_recent_available_dt.strftime('%Y-%m-%d %H:%M')})

            sites_data += [site_data]

    df_map = pd.DataFrame(sites_data).set_index('osuked_id').query('output>0')

    return df_map

# Cell
def df_to_gdf(df, lat_col='latitude', lon_col='longitude'):
    geometry = gpd.points_from_xy(df[lon_col], df[lat_col])
    gdf = gpd.GeoDataFrame(df, geometry=geometry)

    return gdf

# Cell
def generate_map_js(gdf, js_template_fp='templates/map.js', js_docs_fp='docs/js/map.js'):
    geojson = json.loads(gdf.to_json())

    rendered_map_js = Template(open(js_template_fp).read()).render(
        zoom=5,
        center=[53.96, -3.22],
        geojson_features=str(geojson).replace('None', 'null')
    )

    with open(js_docs_fp, 'w', encoding='utf8') as fp:
        fp.write(rendered_map_js)

    return

# Cell
def generate_map_md(most_recent_available_dt, md_template_fp='templates/map.md', md_docs_fp='docs/map.md'):
    update_date = pd.to_datetime(most_recent_available_dt).strftime('%Y-%m-%d %H:%M')
    rendered_map_md = Template(open(md_template_fp).read()).render({'update_date': update_date})

    with open(md_docs_fp, 'w', encoding='utf8') as fp:
        fp.write(rendered_map_md)

    return

# Cell
app = typer.Typer()

# Cell
@app.command()
def generate_map(
    api_key: str=None,
    powerdict_url: str='https://raw.githubusercontent.com/OSUKED/Power-Station-Dictionary/main/data/output/power_stations.csv',
    js_template_fp: str='templates/map.js',
    js_docs_fp: str='docs/js/map.js',
    md_template_fp: str='templates/map.md',
    md_docs_fp: str='docs/map.md'
):
    if api_key is None:
        assert 'BMRS_API_KEY' in os.environ.keys(), 'If the `api_key` is not specified during client initialisation then it must be set to as the environment variable `BMRS_API_KEY`'
        api_key = os.environ['BMRS_API_KEY']

    df_PN = get_latest_PN_df(api_key)

    nearest_half_hour = (pd.Timestamp.now().tz_localize('Europe/London')+pd.Timedelta(minutes=45)).round('30min')
    most_recent_available_dt = max(df_PN.index[pd.to_datetime(df_PN.index).tz_localize('Europe/London')<=nearest_half_hour])
    s_PN = df_PN.loc[most_recent_available_dt]

    df_powerdict = pd.read_csv(powerdict_url)

    osuked_id_mappings = construct_osuked_id_mappings(df_powerdict)
    osuked_id_to_bmu_ids, osuked_id_to_fuel_type, osuked_id_to_name, osuked_id_to_lat_lon = osuked_id_mappings.values()

    df_map = construct_map_df(s_PN, most_recent_available_dt, osuked_id_to_bmu_ids, osuked_id_to_lat_lon, osuked_id_to_fuel_type, osuked_id_to_name)
    gdf_map = df_to_gdf(df_map)

    generate_map_js(gdf_map, js_template_fp=js_template_fp, js_docs_fp=js_docs_fp)
    generate_map_md(most_recent_available_dt, md_template_fp=md_template_fp, md_docs_fp=md_docs_fp)

    return

# Cell
if __name__ == '__main__' and '__file__' in globals():
    app()


# AUTOGENERATED! DO NOT EDIT! File to edit: nbs/07-cli-rebuild.ipynb (unless otherwise specified).

__all__ = ['app', 'rebuild_library']

# Cell
import typer
import shutil
import pandas as pd
from fastcore.foundation import Config

from .dev import nbdev, specgen, rawgen, clientgen

# Cell
app = typer.Typer()

# Cell
@app.command()
def rebuild_library():
    lib_path = str(Config().path('lib_path'))
    dir_root = f'{lib_path}/..'
    endpoints_fp = f'{dir_root}/data/endpoints.csv'

    shutil.rmtree(lib_path)
    nbdev.prepare_nbdev_module()
    nbdev.notebook2script()

    df_endpoints = specgen.load_endpoints_df(endpoints_fp)
    API_spec = specgen.construct_spec(df_endpoints)

    specgen.save_spec(
        API_spec,
        in_fp=f'{dir_root}/templates/open_api_spec.yaml',
        out_fp=f'{dir_root}/data/BMRS_API.yaml'
    )

    rawgen.save_methods(
        functions=rawgen.construct_all_functions(specgen.load_API_yaml(fp=f'{dir_root}/data/BMRS_API.yaml')),
        in_fp=f'{dir_root}/templates/raw_methods.py',
        out_fp=f'{dir_root}/ElexonDataPortal/dev/raw.py'
    )

    clientgen.save_api(
        API_yaml_fp=f'{dir_root}/data/BMRS_API.yaml',
        in_fp=f'{dir_root}/templates/api.py',
        out_fp=f'{dir_root}/ElexonDataPortal/api.py'
    )

    nbdev.add_extra_code_desc_to_mod()

    return

# Cell
if __name__ == '__main__' and '__file__' in globals():
    app()


# AUTOGENERATED! DO NOT EDIT! File to edit: nbs/06-client-gen.ipynb (unless otherwise specified).

__all__ = ['generate_streams', 'save_api', 'generate_streams', 'save_api']

# Cell
from jinja2 import Template

from . import specgen, rawgen, clientprep

# Cell
def generate_streams(API_yaml_fp):
    request_type_to_date_range_example = {
        'SP_and_date': ('2020-01-01', '2020-01-01 1:30'),
        'date_range': ('2020-01-01', '2020-01-07'),
        'date_time_range': ('2020-01-01', '2020-01-07'),
        'year': ('2019-01-01', '2021-01-01'),
        'year_and_month': ('2020-01-01', '2020-06-01'),
        'year_and_week': ('2020-01-01', '2020-06-01'),
        'non_temporal': (None, None)
    }

    API_yaml = specgen.load_API_yaml(API_yaml_fp)
    functions = rawgen.construct_all_functions(API_yaml)
    method_info = clientprep.construct_method_info_dict(API_yaml_fp)

    streams = list()

    for function in functions:
        name = function['name']
        function_method_info = method_info[name]

        stream = dict()
        stream['name'] = name
        stream['description'] = function['description']
        stream['date_range_example'] = request_type_to_date_range_example[function_method_info['request_type']]
        stream['extra_kwargs'] = [param for param in function['parameters'] if param['name'] not in list(function_method_info['kwargs_map'].values())+['APIKey', 'ServiceType']]
        stream['request_type'] = function_method_info['request_type']
        stream['kwargs_map'] = function_method_info['kwargs_map']
        stream['func_params'] = list(function_method_info['func_kwargs'].keys())

        streams += [stream]

    return streams

# Cell
def save_api(
    API_yaml_fp: str,
    in_fp: str='../templates/api.py',
    out_fp: str='../ElexonDataPortal/api.py'
):
    streams = generate_streams(API_yaml_fp)
    rendered_schema = Template(open(in_fp).read()).render(streams=streams)

    with open(out_fp, 'w') as f:
        try:
            f.write(rendered_schema)
        except e as exc:
            raise exc

# Cell
from jinja2 import Template

from . import specgen, rawgen, clientprep

# Cell
def generate_streams(API_yaml_fp):
    request_type_to_date_range_example = {
        'SP_and_date': ('2020-01-01', '2020-01-01 1:30'),
        'date_range': ('2020-01-01', '2020-01-07'),
        'date_time_range': ('2020-01-01', '2020-01-07'),
        'year': ('2019-01-01', '2021-01-01'),
        'year_and_month': ('2020-01-01', '2020-06-01'),
        'year_and_week': ('2020-01-01', '2020-06-01'),
        'non_temporal': (None, None)
    }

    API_yaml = specgen.load_API_yaml(API_yaml_fp)
    functions = rawgen.construct_all_functions(API_yaml)
    method_info = clientprep.construct_method_info_dict(API_yaml_fp)

    streams = list()

    for function in functions:
        name = function['name']
        function_method_info = method_info[name]

        stream = dict()
        stream['name'] = name
        stream['description'] = function['description']
        stream['date_range_example'] = request_type_to_date_range_example[function_method_info['request_type']]
        stream['extra_kwargs'] = [param for param in function['parameters'] if param['name'] not in list(function_method_info['kwargs_map'].values())+['APIKey', 'ServiceType']]
        stream['request_type'] = function_method_info['request_type']
        stream['kwargs_map'] = function_method_info['kwargs_map']
        stream['func_params'] = list(function_method_info['func_kwargs'].keys())

        streams += [stream]

    return streams

# Cell
def save_api(
    API_yaml_fp: str,
    in_fp: str='../templates/api.py',
    out_fp: str='../ElexonDataPortal/api.py'
):
    streams = generate_streams(API_yaml_fp)
    rendered_schema = Template(open(in_fp).read()).render(streams=streams)

    with open(out_fp, 'w') as f:
        try:
            f.write(rendered_schema)
        except e as exc:
            raise exc


import requests


def get_B0610(
    APIKey='AP8DA23',
    SettlementDate='2021-01-01',
    Period='1',
    ServiceType='csv',
    endpoint='https://api.bmreports.com/BMRS/B0610/v1'
):
    """Actual Total Load per Bidding Zone
    """
    
    params = { 
        'APIKey': APIKey,
        'SettlementDate': SettlementDate,
        'Period': Period,
        'ServiceType': ServiceType,
    }
    
    r = requests.get(endpoint, params=params)
    
    return r


def get_B0620(
    APIKey='AP8DA23',
    SettlementDate='2021-01-01',
    Period='1',
    ServiceType='csv',
    endpoint='https://api.bmreports.com/BMRS/B0620/v1'
):
    """Day-Ahead Total Load Forecast per Bidding Zone
    """
    
    params = { 
        'APIKey': APIKey,
        'SettlementDate': SettlementDate,
        'Period': Period,
        'ServiceType': ServiceType,
    }
    
    r = requests.get(endpoint, params=params)
    
    return r


def get_B0630(
    APIKey='AP8DA23',
    Year='2021',
    Week='22',
    ServiceType='csv',
    endpoint='https://api.bmreports.com/BMRS/B0630/v1'
):
    """Week-Ahead Total Load Forecast per Bidding Zone
    """
    
    params = { 
        'APIKey': APIKey,
        'Year': Year,
        'Week': Week,
        'ServiceType': ServiceType,
    }
    
    r = requests.get(endpoint, params=params)
    
    return r


def get_B0640(
    APIKey='AP8DA23',
    Year='2021',
    Month='MAR',
    ServiceType='csv/xml/CSV/XML',
    endpoint='https://api.bmreports.com/BMRS/B0640/v1'
):
    """Month-Ahead Total Load Forecast Per Bidding Zone
    """
    
    params = { 
        'APIKey': APIKey,
        'Year': Year,
        'Month': Month,
        'ServiceType': ServiceType,
    }
    
    r = requests.get(endpoint, params=params)
    
    return r


def get_B0650(
    APIKey='AP8DA23',
    Year='2021',
    ServiceType='csv',
    endpoint='https://api.bmreports.com/BMRS/B0650/v1'
):
    """Year Ahead Total Load Forecast per Bidding Zone
    """
    
    params = { 
        'APIKey': APIKey,
        'Year': Year,
        'ServiceType': ServiceType,
    }
    
    r = requests.get(endpoint, params=params)
    
    return r


def get_B0710(
    APIKey='AP8DA23',
    EndTime='23:59:59',
    StartTime='00:00:00',
    StartDate='2021-01-01',
    EndDate='2021-01-02',
    ServiceType='csv',
    endpoint='https://api.bmreports.com/BMRS/B0710/v1'
):
    """Planned Unavailability of Consumption Units
    """
    
    params = { 
        'APIKey': APIKey,
        'EndTime': EndTime,
        'StartTime': StartTime,
        'StartDate': StartDate,
        'EndDate': EndDate,
        'ServiceType': ServiceType,
    }
    
    r = requests.get(endpoint, params=params)
    
    return r


def get_B0720(
    APIKey='AP8DA23',
    StartDate='2021-01-01',
    StartTime='00:00:00',
    EndDate='2021-01-02',
    EndTime='23:59:59',
    ServiceType='csv',
    endpoint='https://api.bmreports.com/BMRS/B0720/v1'
):
    """Changes In Actual Availability Of Consumption Units
    """
    
    params = { 
        'APIKey': APIKey,
        'StartDate': StartDate,
        'StartTime': StartTime,
        'EndDate': EndDate,
        'EndTime': EndTime,
        'ServiceType': ServiceType,
    }
    
    r = requests.get(endpoint, params=params)
    
    return r


def get_B0810(
    APIKey='AP8DA23',
    Year='2021',
    ServiceType='csv',
    endpoint='https://api.bmreports.com/BMRS/B0810/v1'
):
    """Year Ahead Forecast Margin
    """
    
    params = { 
        'APIKey': APIKey,
        'Year': Year,
        'ServiceType': ServiceType,
    }
    
    r = requests.get(endpoint, params=params)
    
    return r


def get_B0910(
    APIKey='AP8DA23',
    Year='2021',
    ServiceType='csv',
    endpoint='https://api.bmreports.com/BMRS/B0910/v1'
):
    """Expansion and Dismantling Projects
    """
    
    params = { 
        'APIKey': APIKey,
        'Year': Year,
        'ServiceType': ServiceType,
    }
    
    r = requests.get(endpoint, params=params)
    
    return r


def get_B1010(
    APIKey='AP8DA23',
    StartDate='2021-01-01',
    EndDate='2021-01-02',
    StartTime='00:00:00',
    EndTime='23:59:59',
    ServiceType='csv',
    endpoint='https://api.bmreports.com/BMRS/B1010/v1'
):
    """Planned Unavailability In The Transmission Grid
    """
    
    params = { 
        'APIKey': APIKey,
        'StartDate': StartDate,
        'EndDate': EndDate,
        'StartTime': StartTime,
        'EndTime': EndTime,
        'ServiceType': ServiceType,
    }
    
    r = requests.get(endpoint, params=params)
    
    return r


def get_B1020(
    APIKey='AP8DA23',
    StartDate='2021-01-01',
    EndDate='2021-01-02',
    StartTime='00:00:00',
    EndTime='23:59:59',
    ServiceType='csv',
    endpoint='https://api.bmreports.com/BMRS/B1020/v1'
):
    """Changes In Actual Availability In The Transmission Grid
    """
    
    params = { 
        'APIKey': APIKey,
        'StartDate': StartDate,
        'EndDate': EndDate,
        'StartTime': StartTime,
        'EndTime': EndTime,
        'ServiceType': ServiceType,
    }
    
    r = requests.get(endpoint, params=params)
    
    return r


def get_B1030(
    APIKey='AP8DA23',
    StartDate='2021-01-01',
    EndDate='2021-01-02',
    StartTime='00:00:00',
    EndTime='23:59:59',
    ServiceType='csv',
    endpoint='https://api.bmreports.com/BMRS/B1030/v1'
):
    """Changes In Actual Availability of Offshore Grid Infrastructure
    """
    
    params = { 
        'APIKey': APIKey,
        'StartDate': StartDate,
        'EndDate': EndDate,
        'StartTime': StartTime,
        'EndTime': EndTime,
        'ServiceType': ServiceType,
    }
    
    r = requests.get(endpoint, params=params)
    
    return r


def get_B1320(
    APIKey='AP8DA23',
    SettlementDate='2021-01-01',
    Period='1',
    ServiceType='csv',
    endpoint='https://api.bmreports.com/BMRS/B1320/v1'
):
    """Congestion Management Measures Countertrading
    """
    
    params = { 
        'APIKey': APIKey,
        'SettlementDate': SettlementDate,
        'Period': Period,
        'ServiceType': ServiceType,
    }
    
    r = requests.get(endpoint, params=params)
    
    return r


def get_B1330(
    APIKey='AP8DA23',
    Year='2021',
    Month='Mar',
    ServiceType='csv',
    endpoint='https://api.bmreports.com/BMRS/B1330/v1'
):
    """Congestion Management Measures Costs of Congestion Management
    """
    
    params = { 
        'APIKey': APIKey,
        'Year': Year,
        'Month': Month,
        'ServiceType': ServiceType,
    }
    
    r = requests.get(endpoint, params=params)
    
    return r


def get_B1410(
    APIKey='AP8DA23',
    Year='2021',
    ServiceType='csv',
    endpoint='https://api.bmreports.com/BMRS/B1410/v1'
):
    """Installed Generation Capacity Aggregated
    """
    
    params = { 
        'APIKey': APIKey,
        'Year': Year,
        'ServiceType': ServiceType,
    }
    
    r = requests.get(endpoint, params=params)
    
    return r


def get_B1420(
    APIKey='AP8DA23',
    Year='2021',
    ServiceType='csv',
    endpoint='https://api.bmreports.com/BMRS/B1420/v1'
):
    """Installed Generation Capacity per Unit
    """
    
    params = { 
        'APIKey': APIKey,
        'Year': Year,
        'ServiceType': ServiceType,
    }
    
    r = requests.get(endpoint, params=params)
    
    return r


def get_B1430(
    APIKey='AP8DA23',
    SettlementDate='2021-01-01',
    Period='1',
    ServiceType='csv',
    endpoint='https://api.bmreports.com/BMRS/B1430/v1'
):
    """Day-Ahead Aggregated Generation
    """
    
    params = { 
        'APIKey': APIKey,
        'SettlementDate': SettlementDate,
        'Period': Period,
        'ServiceType': ServiceType,
    }
    
    r = requests.get(endpoint, params=params)
    
    return r


def get_B1440(
    APIKey='AP8DA23',
    SettlementDate='2021-01-01',
    Period='1',
    ProcessType='Day Ahead',
    ServiceType='csv',
    endpoint='https://api.bmreports.com/BMRS/B1440/v1'
):
    """Generation forecasts for Wind and Solar
    """
    
    params = { 
        'APIKey': APIKey,
        'SettlementDate': SettlementDate,
        'Period': Period,
        'ProcessType': ProcessType,
        'ServiceType': ServiceType,
    }
    
    r = requests.get(endpoint, params=params)
    
    return r


def get_B1510(
    APIKey='AP8DA23',
    StartDate='2021-01-01',
    EndDate='2021-01-02',
    StartTime='00:00:00',
    EndTime='23:59:59',
    ServiceType='csv',
    endpoint='https://api.bmreports.com/BMRS/B1510/v1'
):
    """Planned Unavailability of Generation Units
    """
    
    params = { 
        'APIKey': APIKey,
        'StartDate': StartDate,
        'EndDate': EndDate,
        'StartTime': StartTime,
        'EndTime': EndTime,
        'ServiceType': ServiceType,
    }
    
    r = requests.get(endpoint, params=params)
    
    return r


def get_B1520(
    APIKey='AP8DA23',
    StartDate='2021-01-01',
    EndDate='2021-01-02',
    StartTime='00:00:00',
    EndTime='23:59:59',
    ServiceType='csv',
    endpoint='https://api.bmreports.com/BMRS/B1520/v1'
):
    """Changes In Actual Availability of Generation Units
    """
    
    params = { 
        'APIKey': APIKey,
        'StartDate': StartDate,
        'EndDate': EndDate,
        'StartTime': StartTime,
        'EndTime': EndTime,
        'ServiceType': ServiceType,
    }
    
    r = requests.get(endpoint, params=params)
    
    return r


def get_B1530(
    APIKey='AP8DA23',
    StartDate='2021-01-01',
    EndDate='2021-01-02',
    StartTime='00:00:00',
    EndTime='23:59:59',
    ServiceType='csv',
    endpoint='https://api.bmreports.com/BMRS/B1530/v1'
):
    """Planned Unavailability of Production Units
    """
    
    params = { 
        'APIKey': APIKey,
        'StartDate': StartDate,
        'EndDate': EndDate,
        'StartTime': StartTime,
        'EndTime': EndTime,
        'ServiceType': ServiceType,
    }
    
    r = requests.get(endpoint, params=params)
    
    return r


def get_B1540(
    APIKey='AP8DA23',
    StartDate='2021-01-01',
    EndDate='2021-01-02',
    StartTime='00:00:00',
    EndTime='23:59:59',
    ServiceType='csv',
    endpoint='https://api.bmreports.com/BMRS/B1540/v1'
):
    """Changes In Actual Availability of Production Units
    """
    
    params = { 
        'APIKey': APIKey,
        'StartDate': StartDate,
        'EndDate': EndDate,
        'StartTime': StartTime,
        'EndTime': EndTime,
        'ServiceType': ServiceType,
    }
    
    r = requests.get(endpoint, params=params)
    
    return r


def get_B1610(
    APIKey='AP8DA23',
    SettlementDate='2021-01-01',
    Period='1',
    NGCBMUnitID='*',
    ServiceType='csv',
    endpoint='https://api.bmreports.com/BMRS/B1610/v2'
):
    """Actual Generation Output per Generation Unit
    """
    
    params = { 
        'APIKey': APIKey,
        'SettlementDate': SettlementDate,
        'Period': Period,
        'NGCBMUnitID': NGCBMUnitID,
        'ServiceType': ServiceType,
    }
    
    r = requests.get(endpoint, params=params)
    
    return r


def get_B1620(
    APIKey='AP8DA23',
    SettlementDate='2021-01-01',
    Period='1',
    ServiceType='csv',
    endpoint='https://api.bmreports.com/BMRS/B1620/v1'
):
    """Actual Aggregated Generation per Type
    """
    
    params = { 
        'APIKey': APIKey,
        'SettlementDate': SettlementDate,
        'Period': Period,
        'ServiceType': ServiceType,
    }
    
    r = requests.get(endpoint, params=params)
    
    return r


def get_B1630(
    APIKey='AP8DA23',
    SettlementDate='2021-01-01',
    Period='1',
    ServiceType='csv',
    endpoint='https://api.bmreports.com/BMRS/B1630/v1'
):
    """Actual Or Estimated Wind and Solar Power Generation
    """
    
    params = { 
        'APIKey': APIKey,
        'SettlementDate': SettlementDate,
        'Period': Period,
        'ServiceType': ServiceType,
    }
    
    r = requests.get(endpoint, params=params)
    
    return r


def get_B1720(
    APIKey='AP8DA23',
    SettlementDate='2021-01-01',
    Period='1',
    ServiceType='csv',
    endpoint='https://api.bmreports.com/BMRS/B1720/v1'
):
    """Amount Of Balancing Reserves Under Contract Service
    """
    
    params = { 
        'APIKey': APIKey,
        'SettlementDate': SettlementDate,
        'Period': Period,
        'ServiceType': ServiceType,
    }
    
    r = requests.get(endpoint, params=params)
    
    return r


def get_B1730(
    APIKey='AP8DA23',
    SettlementDate='2021-01-01',
    Period='1',
    ServiceType='csv',
    endpoint='https://api.bmreports.com/BMRS/B1730/v1'
):
    """Prices Of Procured Balancing Reserves Service
    """
    
    params = { 
        'APIKey': APIKey,
        'SettlementDate': SettlementDate,
        'Period': Period,
        'ServiceType': ServiceType,
    }
    
    r = requests.get(endpoint, params=params)
    
    return r


def get_B1740(
    APIKey='AP8DA23',
    SettlementDate='2021-01-01',
    Period='1',
    ServiceType='csv',
    endpoint='https://api.bmreports.com/BMRS/B1740/v1'
):
    """Accepted Aggregated Offers
    """
    
    params = { 
        'APIKey': APIKey,
        'SettlementDate': SettlementDate,
        'Period': Period,
        'ServiceType': ServiceType,
    }
    
    r = requests.get(endpoint, params=params)
    
    return r


def get_B1750(
    APIKey='AP8DA23',
    SettlementDate='2021-01-01',
    Period='1',
    ServiceType='csv',
    endpoint='https://api.bmreports.com/BMRS/B1750/v1'
):
    """Activated Balancing Energy
    """
    
    params = { 
        'APIKey': APIKey,
        'SettlementDate': SettlementDate,
        'Period': Period,
        'ServiceType': ServiceType,
    }
    
    r = requests.get(endpoint, params=params)
    
    return r


def get_B1760(
    APIKey='AP8DA23',
    SettlementDate='2021-01-01',
    Period='1',
    ServiceType='csv',
    endpoint='https://api.bmreports.com/BMRS/B1760/v1'
):
    """Prices Of Activated Balancing Energy
    """
    
    params = { 
        'APIKey': APIKey,
        'SettlementDate': SettlementDate,
        'Period': Period,
        'ServiceType': ServiceType,
    }
    
    r = requests.get(endpoint, params=params)
    
    return r


def get_B1770(
    APIKey='AP8DA23',
    SettlementDate='2021-01-01',
    Period='1',
    ServiceType='csv',
    endpoint='https://api.bmreports.com/BMRS/B1770/v1'
):
    """Imbalance Prices
    """
    
    params = { 
        'APIKey': APIKey,
        'SettlementDate': SettlementDate,
        'Period': Period,
        'ServiceType': ServiceType,
    }
    
    r = requests.get(endpoint, params=params)
    
    return r


def get_B1780(
    APIKey='AP8DA23',
    SettlementDate='2021-01-01',
    Period='1',
    ServiceType='csv',
    endpoint='https://api.bmreports.com/BMRS/B1780/v1'
):
    """Aggregated Imbalance Volumes
    """
    
    params = { 
        'APIKey': APIKey,
        'SettlementDate': SettlementDate,
        'Period': Period,
        'ServiceType': ServiceType,
    }
    
    r = requests.get(endpoint, params=params)
    
    return r


def get_B1790(
    APIKey='AP8DA23',
    Year='2021',
    Month='MAR',
    ServiceType='csv/xml/CSV/XML',
    endpoint='https://api.bmreports.com/BMRS/B1790/v1'
):
    """Financial Expenses and Income For Balancing
    """
    
    params = { 
        'APIKey': APIKey,
        'Year': Year,
        'Month': Month,
        'ServiceType': ServiceType,
    }
    
    r = requests.get(endpoint, params=params)
    
    return r


def get_B1810(
    APIKey='AP8DA23',
    SettlementDate='2021-01-01',
    Period='1',
    ServiceType='csv',
    endpoint='https://api.bmreports.com/BMRS/B1810/v1'
):
    """Cross-Border Balancing Volumes of Exchanged Bids and Offers
    """
    
    params = { 
        'APIKey': APIKey,
        'SettlementDate': SettlementDate,
        'Period': Period,
        'ServiceType': ServiceType,
    }
    
    r = requests.get(endpoint, params=params)
    
    return r


def get_B1820(
    APIKey='AP8DA23',
    SettlementDate='2021-01-01',
    Period='1',
    ServiceType='csv',
    endpoint='https://api.bmreports.com/BMRS/B1820/v1'
):
    """Cross-Border Balancing Prices
    """
    
    params = { 
        'APIKey': APIKey,
        'SettlementDate': SettlementDate,
        'Period': Period,
        'ServiceType': ServiceType,
    }
    
    r = requests.get(endpoint, params=params)
    
    return r


def get_B1830(
    APIKey='AP8DA23',
    SettlementDate='2021-01-01',
    Period='1',
    ServiceType='csv',
    endpoint='https://api.bmreports.com/BMRS/B1830/v1'
):
    """Cross-border Balancing Energy Activated
    """
    
    params = { 
        'APIKey': APIKey,
        'SettlementDate': SettlementDate,
        'Period': Period,
        'ServiceType': ServiceType,
    }
    
    r = requests.get(endpoint, params=params)
    
    return r


def get_BOD(
    APIKey='AP8DA23',
    SettlementDate='2021-01-01',
    SettlementPeriod='12',
    BMUnitId='2__AEENG000, G, E.ON Energy, Solutions Limited, EAS-EST01',
    BMUnitType='G, S, E, I, T, etc',
    LeadPartyName='AES New Energy Limited',
    NGCBMUnit='EAS-ASP01, AES New Energy Limited, G, 2__AAEPD000',
    Name='2__AAEPD000',
    ServiceType='csv/xml/CSV/XML',
    endpoint='https://api.bmreports.com/BMRS/BOD/v1'
):
    """Bid Offer Level Data
    """
    
    params = { 
        'APIKey': APIKey,
        'SettlementDate': SettlementDate,
        'SettlementPeriod': SettlementPeriod,
        'BMUnitId': BMUnitId,
        'BMUnitType': BMUnitType,
        'LeadPartyName': LeadPartyName,
        'NGCBMUnit': NGCBMUnit,
        'Name': Name,
        'ServiceType': ServiceType,
    }
    
    r = requests.get(endpoint, params=params)
    
    return r


def get_CDN(
    APIKey='AP8DA23',
    FromClearedDate='2021-01-01',
    ToClearedDate='2021-01-02',
    ServiceType='xml/XML/csv/CSV',
    endpoint='https://api.bmreports.com/BMRS/CDN/v1'
):
    """Credit Default Notice Data
    """
    
    params = { 
        'APIKey': APIKey,
        'FromClearedDate': FromClearedDate,
        'ToClearedDate': ToClearedDate,
        'ServiceType': ServiceType,
    }
    
    r = requests.get(endpoint, params=params)
    
    return r


def get_DETSYSPRICES(
    APIKey='AP8DA23',
    SettlementDate='2014-01-02',
    SettlementPeriod='2',
    ServiceType='xml/XML/csv/CSV',
    endpoint='https://api.bmreports.com/BMRS/DETSYSPRICES/v1'
):
    """Detailed System Prices
    """
    
    params = { 
        'APIKey': APIKey,
        'SettlementDate': SettlementDate,
        'SettlementPeriod': SettlementPeriod,
        'ServiceType': ServiceType,
    }
    
    r = requests.get(endpoint, params=params)
    
    return r


def get_DEVINDOD(
    APIKey='AP8DA23',
    FromDate='2021-01-01',
    ToDate='2021-01-02',
    ServiceType='xml/XML/csv/CSV',
    endpoint='https://api.bmreports.com/BMRS/DEVINDOD/v1'
):
    """Daily Energy Volume Data
    """
    
    params = { 
        'APIKey': APIKey,
        'FromDate': FromDate,
        'ToDate': ToDate,
        'ServiceType': ServiceType,
    }
    
    r = requests.get(endpoint, params=params)
    
    return r


def get_DISBSAD(
    APIKey='AP8DA23',
    SettlementDate='2021-01-01',
    SettlementPeriod='1',
    ServiceType='xml/XML/csv/CSV',
    endpoint='https://api.bmreports.com/BMRS/DISBSAD/v1'
):
    """Balancing Services Adjustment Action Data
    """
    
    params = { 
        'APIKey': APIKey,
        'SettlementDate': SettlementDate,
        'SettlementPeriod': SettlementPeriod,
        'ServiceType': ServiceType,
    }
    
    r = requests.get(endpoint, params=params)
    
    return r


def get_FORDAYDEM(
    APIKey='AP8DA23',
    ZoneIdentifier='N',
    FromDate='2021-01-01',
    ToDate='2021-01-02',
    ServiceType='csv/xml/CSV/XML',
    endpoint='https://api.bmreports.com/BMRS/FORDAYDEM/v1'
):
    """Forecast Day and Day Ahead Demand Data
    """
    
    params = { 
        'APIKey': APIKey,
        'ZoneIdentifier': ZoneIdentifier,
        'FromDate': FromDate,
        'ToDate': ToDate,
        'ServiceType': ServiceType,
    }
    
    r = requests.get(endpoint, params=params)
    
    return r


def get_FREQ(
    APIKey='AP8DA23',
    FromDateTime='2021-01-01 00:01:00',
    ToDateTime='2021-02-01 23:59:00',
    ServiceType='xml/XML/csv/CSV',
    endpoint='https://api.bmreports.com/BMRS/FREQ/v1'
):
    """Rolling System Frequency
    """
    
    params = { 
        'APIKey': APIKey,
        'FromDateTime': FromDateTime,
        'ToDateTime': ToDateTime,
        'ServiceType': ServiceType,
    }
    
    r = requests.get(endpoint, params=params)
    
    return r


def get_FUELHH(
    APIKey='AP8DA23',
    FromDate='44197',
    ToDate='44228',
    ServiceType='csv/xml/CSV/XML',
    endpoint='https://api.bmreports.com/BMRS/FUELHH/v1'
):
    """Half Hourly Outturn Generation by Fuel Type
    """
    
    params = { 
        'APIKey': APIKey,
        'FromDate': FromDate,
        'ToDate': ToDate,
        'ServiceType': ServiceType,
    }
    
    r = requests.get(endpoint, params=params)
    
    return r


def get_MELIMBALNGC(
    APIKey='AP8DA23',
    ZoneIdentifier='N',
    FromDate='2021-01-01',
    ToDate='2021-01-02',
    ServiceType='csv/xml/CSV/XML',
    endpoint='https://api.bmreports.com/BMRS/MELIMBALNGC/v1'
):
    """Forecast Day and Day Ahead Margin and Imbalance Data
    """
    
    params = { 
        'APIKey': APIKey,
        'ZoneIdentifier': ZoneIdentifier,
        'FromDate': FromDate,
        'ToDate': ToDate,
        'ServiceType': ServiceType,
    }
    
    r = requests.get(endpoint, params=params)
    
    return r


def get_MID(
    APIKey='AP8DA23',
    FromSettlementDate='2021-01-01',
    ToSettlementDate='2021-01-02',
    Period='*',
    ServiceType='csv/CSV/XML/xml',
    endpoint='https://api.bmreports.com/BMRS/MID/v1'
):
    """Market Index Data
    """
    
    params = { 
        'APIKey': APIKey,
        'FromSettlementDate': FromSettlementDate,
        'ToSettlementDate': ToSettlementDate,
        'Period': Period,
        'ServiceType': ServiceType,
    }
    
    r = requests.get(endpoint, params=params)
    
    return r


def get_MessageDetailRetrieval(
    APIKey='AP8DA23',
    MessageId='',
    ParticipantId='',
    SequenceId='',
    ActiveFlag='N',
    ServiceType='xml/XML/csv/CSV.',
    endpoint='https://api.bmreports.com/BMRS/MessageDetailRetrieval/v1'
):
    """REMIT Flow - Message List Retrieval
    """
    
    params = { 
        'APIKey': APIKey,
        'MessageId': MessageId,
        'ParticipantId': ParticipantId,
        'SequenceId': SequenceId,
        'ActiveFlag': ActiveFlag,
        'ServiceType': ServiceType,
    }
    
    r = requests.get(endpoint, params=params)
    
    return r


def get_MessageListRetrieval(
    APIKey='AP8DA23',
    EventStart='2021-01-01',
    EventEnd='2021-01-02',
    PublicationFrom='2021-01-01',
    PublicationTo='2021-01-02',
    ParticipantId='',
    MessageID='',
    AssetID='',
    EventType='',
    FuelType='',
    MessageType='',
    UnavailabilityType='',
    ServiceType='xml/XML/csv/CSV',
    endpoint='https://api.bmreports.com/BMRS/MessageListRetrieval/v1'
):
    """REMIT Flow - Message List Retrieval
    """
    
    params = { 
        'APIKey': APIKey,
        'EventStart': EventStart,
        'EventEnd': EventEnd,
        'PublicationFrom': PublicationFrom,
        'PublicationTo': PublicationTo,
        'ParticipantId': ParticipantId,
        'MessageID': MessageID,
        'AssetID': AssetID,
        'EventType': EventType,
        'FuelType': FuelType,
        'MessageType': MessageType,
        'UnavailabilityType': UnavailabilityType,
        'ServiceType': ServiceType,
    }
    
    r = requests.get(endpoint, params=params)
    
    return r


def get_NETBSAD(
    APIKey='AP8DA23',
    SettlementDate='2021-01-01',
    SettlementPeriod='1',
    isTwoDayWindow='FALSE',
    ServiceType='xml/XML/csv/CSV',
    endpoint='https://api.bmreports.com/BMRS/NETBSAD/v1'
):
    """Balancing Service Adjustment Data
    """
    
    params = { 
        'APIKey': APIKey,
        'SettlementDate': SettlementDate,
        'SettlementPeriod': SettlementPeriod,
        'isTwoDayWindow': isTwoDayWindow,
        'ServiceType': ServiceType,
    }
    
    r = requests.get(endpoint, params=params)
    
    return r


def get_NONBM(
    FromDate='2021-01-01',
    ToDate='2021-01-02',
    APIKey='AP8DA23',
    ServiceType='csv/CSV/XML/xml',
    endpoint='https://api.bmreports.com/BMRS/NONBM/v1'
):
    """Non BM STOR Instructed Volume Data
    """
    
    params = { 
        'FromDate': FromDate,
        'ToDate': ToDate,
        'APIKey': APIKey,
        'ServiceType': ServiceType,
    }
    
    r = requests.get(endpoint, params=params)
    
    return r


def get_PHYBMDATA(
    APIKey='AP8DA23',
    SettlementDate='2021-01-01',
    SettlementPeriod='12',
    BMUnitId='',
    BMUnitType='',
    LeadPartyName='',
    NGCBMUnit='',
    Name='',
    ServiceType='csv/xml/CSV/XML',
    endpoint='https://api.bmreports.com/BMRS/PHYBMDATA/v1'
):
    """Physical Data
    """
    
    params = { 
        'APIKey': APIKey,
        'SettlementDate': SettlementDate,
        'SettlementPeriod': SettlementPeriod,
        'BMUnitId': BMUnitId,
        'BMUnitType': BMUnitType,
        'LeadPartyName': LeadPartyName,
        'NGCBMUnit': NGCBMUnit,
        'Name': Name,
        'ServiceType': ServiceType,
    }
    
    r = requests.get(endpoint, params=params)
    
    return r


def get_SYSDEM(
    APIKey='AP8DA23',
    FromDate='2021-01-01',
    ToDate='2021-01-02',
    ServiceType='csv/xml/CSV/XML',
    endpoint='https://api.bmreports.com/BMRS/SYSDEM/v1'
):
    """System Demand
    """
    
    params = { 
        'APIKey': APIKey,
        'FromDate': FromDate,
        'ToDate': ToDate,
        'ServiceType': ServiceType,
    }
    
    r = requests.get(endpoint, params=params)
    
    return r


def get_SYSWARN(
    APIKey='AP8DA23',
    FromDate='2021-01-01',
    ToDate='2021-01-02',
    ServiceType='csv/CSV/xml/XML',
    endpoint='https://api.bmreports.com/BMRS/SYSWARN/v1'
):
    """System Warnings
    """
    
    params = { 
        'APIKey': APIKey,
        'FromDate': FromDate,
        'ToDate': ToDate,
        'ServiceType': ServiceType,
    }
    
    r = requests.get(endpoint, params=params)
    
    return r


def get_TEMP(
    APIKey='AP8DA23',
    FromDate='2021-01-01',
    ToDate='2021-01-02',
    ServiceType='csv/CSV/xml/XML',
    endpoint='https://api.bmreports.com/BMRS/TEMP/v1'
):
    """Temperature Data
    """
    
    params = { 
        'APIKey': APIKey,
        'FromDate': FromDate,
        'ToDate': ToDate,
        'ServiceType': ServiceType,
    }
    
    r = requests.get(endpoint, params=params)
    
    return r


def get_WINDFORFUELHH(
    APIKey='AP8DA23',
    FromDate='2021-01-01',
    ToDate='2021-01-02',
    ServiceType='csv/xml/CSV/XML',
    endpoint='https://api.bmreports.com/BMRS/WINDFORFUELHH/v1'
):
    """Wind Generation Forecast and Out-turn Data
    """
    
    params = { 
        'APIKey': APIKey,
        'FromDate': FromDate,
        'ToDate': ToDate,
        'ServiceType': ServiceType,
    }
    
    r = requests.get(endpoint, params=params)
    
    return r




import os
from ElexonDataPortal.dev import orchestrator

class Client:
    def __init__(self, api_key: str=None):
        if api_key is None:
            assert 'BMRS_API_KEY' in os.environ.keys(), 'If the `api_key` is not specified during client initialisation then it must be set to as the environment variable `BMRS_API_KEY`'
            api_key = os.environ['BMRS_API_KEY']
            
        self.api_key = api_key
        self.set_method_descs()
        
        
    def set_method_descs(self):
        get_methods_names = [attr for attr in dir(self) if attr[:4]=='get_']
        get_method_descs = [getattr(self, get_methods_name).__doc__.split('\n')[1].strip() for get_methods_name in get_methods_names]

        self.methods = dict(zip(get_methods_names, get_method_descs))
        
    
    def get_B0610(
        self,
        start_date: str='2020-01-01', 
        end_date: str='2020-01-01 1:30',
    ):
        """
        Actual Total Load per Bidding Zone
        
        Parameters:
            start_date (str)
            end_date (str)
        """
        
        df = orchestrator.query_orchestrator(
            method='get_B0610',
            api_key=self.api_key,
            request_type='SP_and_date',
            kwargs_map={'date': 'SettlementDate', 'SP': 'Period'},
            func_params=['APIKey', 'date', 'SP', 'ServiceType'],
            start_date=start_date,
            end_date=end_date,
        )
        
        return df
    
    
    def get_B0620(
        self,
        start_date: str='2020-01-01', 
        end_date: str='2020-01-01 1:30',
    ):
        """
        Day-Ahead Total Load Forecast per Bidding Zone
        
        Parameters:
            start_date (str)
            end_date (str)
        """
        
        df = orchestrator.query_orchestrator(
            method='get_B0620',
            api_key=self.api_key,
            request_type='SP_and_date',
            kwargs_map={'date': 'SettlementDate', 'SP': 'Period'},
            func_params=['APIKey', 'date', 'SP', 'ServiceType'],
            start_date=start_date,
            end_date=end_date,
        )
        
        return df
    
    
    def get_B0630(
        self,
        start_date: str='2020-01-01', 
        end_date: str='2020-06-01',
    ):
        """
        Week-Ahead Total Load Forecast per Bidding Zone
        
        Parameters:
            start_date (str)
            end_date (str)
        """
        
        df = orchestrator.query_orchestrator(
            method='get_B0630',
            api_key=self.api_key,
            request_type='year_and_week',
            kwargs_map={'year': 'Year', 'week': 'Week'},
            func_params=['APIKey', 'year', 'week', 'ServiceType'],
            start_date=start_date,
            end_date=end_date,
        )
        
        return df
    
    
    def get_B0640(
        self,
        start_date: str='2020-01-01', 
        end_date: str='2020-06-01',
    ):
        """
        Month-Ahead Total Load Forecast Per Bidding Zone
        
        Parameters:
            start_date (str)
            end_date (str)
        """
        
        df = orchestrator.query_orchestrator(
            method='get_B0640',
            api_key=self.api_key,
            request_type='year_and_month',
            kwargs_map={'year': 'Year', 'month': 'Month'},
            func_params=['APIKey', 'year', 'month', 'ServiceType'],
            start_date=start_date,
            end_date=end_date,
        )
        
        return df
    
    
    def get_B0650(
        self,
        start_date: str='2019-01-01', 
        end_date: str='2021-01-01',
    ):
        """
        Year Ahead Total Load Forecast per Bidding Zone
        
        Parameters:
            start_date (str)
            end_date (str)
        """
        
        df = orchestrator.query_orchestrator(
            method='get_B0650',
            api_key=self.api_key,
            request_type='year',
            kwargs_map={'year': 'Year'},
            func_params=['APIKey', 'year', 'ServiceType'],
            start_date=start_date,
            end_date=end_date,
        )
        
        return df
    
    
    def get_B0710(
        self,
        start_date: str='2020-01-01', 
        end_date: str='2020-01-07',
    ):
        """
        Planned Unavailability of Consumption Units
        
        Parameters:
            start_date (str)
            end_date (str)
        """
        
        df = orchestrator.query_orchestrator(
            method='get_B0710',
            api_key=self.api_key,
            request_type='date_time_range',
            kwargs_map={'end_time': 'EndTime', 'start_time': 'StartTime', 'start_date': 'StartDate', 'end_date': 'EndDate'},
            func_params=['APIKey', 'end_time', 'start_time', 'start_date', 'end_date', 'ServiceType'],
            start_date=start_date,
            end_date=end_date,
        )
        
        return df
    
    
    def get_B0720(
        self,
        start_date: str='2020-01-01', 
        end_date: str='2020-01-07',
    ):
        """
        Changes In Actual Availability Of Consumption Units
        
        Parameters:
            start_date (str)
            end_date (str)
        """
        
        df = orchestrator.query_orchestrator(
            method='get_B0720',
            api_key=self.api_key,
            request_type='date_time_range',
            kwargs_map={'start_date': 'StartDate', 'start_time': 'StartTime', 'end_date': 'EndDate', 'end_time': 'EndTime'},
            func_params=['APIKey', 'start_date', 'start_time', 'end_date', 'end_time', 'ServiceType'],
            start_date=start_date,
            end_date=end_date,
        )
        
        return df
    
    
    def get_B0810(
        self,
        start_date: str='2019-01-01', 
        end_date: str='2021-01-01',
    ):
        """
        Year Ahead Forecast Margin
        
        Parameters:
            start_date (str)
            end_date (str)
        """
        
        df = orchestrator.query_orchestrator(
            method='get_B0810',
            api_key=self.api_key,
            request_type='year',
            kwargs_map={'year': 'Year'},
            func_params=['APIKey', 'year', 'ServiceType'],
            start_date=start_date,
            end_date=end_date,
        )
        
        return df
    
    
    def get_B0910(
        self,
        start_date: str='2019-01-01', 
        end_date: str='2021-01-01',
    ):
        """
        Expansion and Dismantling Projects
        
        Parameters:
            start_date (str)
            end_date (str)
        """
        
        df = orchestrator.query_orchestrator(
            method='get_B0910',
            api_key=self.api_key,
            request_type='year',
            kwargs_map={'year': 'Year'},
            func_params=['APIKey', 'year', 'ServiceType'],
            start_date=start_date,
            end_date=end_date,
        )
        
        return df
    
    
    def get_B1010(
        self,
        start_date: str='2020-01-01', 
        end_date: str='2020-01-07',
    ):
        """
        Planned Unavailability In The Transmission Grid
        
        Parameters:
            start_date (str)
            end_date (str)
        """
        
        df = orchestrator.query_orchestrator(
            method='get_B1010',
            api_key=self.api_key,
            request_type='date_time_range',
            kwargs_map={'start_date': 'StartDate', 'end_date': 'EndDate', 'start_time': 'StartTime', 'end_time': 'EndTime'},
            func_params=['APIKey', 'start_date', 'end_date', 'start_time', 'end_time', 'ServiceType'],
            start_date=start_date,
            end_date=end_date,
        )
        
        return df
    
    
    def get_B1020(
        self,
        start_date: str='2020-01-01', 
        end_date: str='2020-01-07',
    ):
        """
        Changes In Actual Availability In The Transmission Grid
        
        Parameters:
            start_date (str)
            end_date (str)
        """
        
        df = orchestrator.query_orchestrator(
            method='get_B1020',
            api_key=self.api_key,
            request_type='date_time_range',
            kwargs_map={'start_date': 'StartDate', 'end_date': 'EndDate', 'start_time': 'StartTime', 'end_time': 'EndTime'},
            func_params=['APIKey', 'start_date', 'end_date', 'start_time', 'end_time', 'ServiceType'],
            start_date=start_date,
            end_date=end_date,
        )
        
        return df
    
    
    def get_B1030(
        self,
        start_date: str='2020-01-01', 
        end_date: str='2020-01-07',
    ):
        """
        Changes In Actual Availability of Offshore Grid Infrastructure
        
        Parameters:
            start_date (str)
            end_date (str)
        """
        
        df = orchestrator.query_orchestrator(
            method='get_B1030',
            api_key=self.api_key,
            request_type='date_time_range',
            kwargs_map={'start_date': 'StartDate', 'end_date': 'EndDate', 'start_time': 'StartTime', 'end_time': 'EndTime'},
            func_params=['APIKey', 'start_date', 'end_date', 'start_time', 'end_time', 'ServiceType'],
            start_date=start_date,
            end_date=end_date,
        )
        
        return df
    
    
    def get_B1320(
        self,
        start_date: str='2020-01-01', 
        end_date: str='2020-01-01 1:30',
    ):
        """
        Congestion Management Measures Countertrading
        
        Parameters:
            start_date (str)
            end_date (str)
        """
        
        df = orchestrator.query_orchestrator(
            method='get_B1320',
            api_key=self.api_key,
            request_type='SP_and_date',
            kwargs_map={'date': 'SettlementDate', 'SP': 'Period'},
            func_params=['APIKey', 'date', 'SP', 'ServiceType'],
            start_date=start_date,
            end_date=end_date,
        )
        
        return df
    
    
    def get_B1330(
        self,
        start_date: str='2020-01-01', 
        end_date: str='2020-06-01',
    ):
        """
        Congestion Management Measures Costs of Congestion Management
        
        Parameters:
            start_date (str)
            end_date (str)
        """
        
        df = orchestrator.query_orchestrator(
            method='get_B1330',
            api_key=self.api_key,
            request_type='year_and_month',
            kwargs_map={'year': 'Year', 'month': 'Month'},
            func_params=['APIKey', 'year', 'month', 'ServiceType'],
            start_date=start_date,
            end_date=end_date,
        )
        
        return df
    
    
    def get_B1410(
        self,
        start_date: str='2019-01-01', 
        end_date: str='2021-01-01',
    ):
        """
        Installed Generation Capacity Aggregated
        
        Parameters:
            start_date (str)
            end_date (str)
        """
        
        df = orchestrator.query_orchestrator(
            method='get_B1410',
            api_key=self.api_key,
            request_type='year',
            kwargs_map={'year': 'Year'},
            func_params=['APIKey', 'year', 'ServiceType'],
            start_date=start_date,
            end_date=end_date,
        )
        
        return df
    
    
    def get_B1420(
        self,
        start_date: str='2019-01-01', 
        end_date: str='2021-01-01',
    ):
        """
        Installed Generation Capacity per Unit
        
        Parameters:
            start_date (str)
            end_date (str)
        """
        
        df = orchestrator.query_orchestrator(
            method='get_B1420',
            api_key=self.api_key,
            request_type='year',
            kwargs_map={'year': 'Year'},
            func_params=['APIKey', 'year', 'ServiceType'],
            start_date=start_date,
            end_date=end_date,
        )
        
        return df
    
    
    def get_B1430(
        self,
        start_date: str='2020-01-01', 
        end_date: str='2020-01-01 1:30',
    ):
        """
        Day-Ahead Aggregated Generation
        
        Parameters:
            start_date (str)
            end_date (str)
        """
        
        df = orchestrator.query_orchestrator(
            method='get_B1430',
            api_key=self.api_key,
            request_type='SP_and_date',
            kwargs_map={'date': 'SettlementDate', 'SP': 'Period'},
            func_params=['APIKey', 'date', 'SP', 'ServiceType'],
            start_date=start_date,
            end_date=end_date,
        )
        
        return df
    
    
    def get_B1440(
        self,
        start_date: str='2020-01-01', 
        end_date: str='2020-01-01 1:30',
        ProcessType: str='Day Ahead',
    ):
        """
        Generation forecasts for Wind and Solar
        
        Parameters:
            start_date (str)
            end_date (str)
            ProcessType (str)
        """
        
        df = orchestrator.query_orchestrator(
            method='get_B1440',
            api_key=self.api_key,
            request_type='SP_and_date',
            kwargs_map={'date': 'SettlementDate', 'SP': 'Period'},
            func_params=['APIKey', 'date', 'SP', 'ProcessType', 'ServiceType'],
            start_date=start_date,
            end_date=end_date,
            ProcessType=ProcessType,
        )
        
        return df
    
    
    def get_B1510(
        self,
        start_date: str='2020-01-01', 
        end_date: str='2020-01-07',
    ):
        """
        Planned Unavailability of Generation Units
        
        Parameters:
            start_date (str)
            end_date (str)
        """
        
        df = orchestrator.query_orchestrator(
            method='get_B1510',
            api_key=self.api_key,
            request_type='date_time_range',
            kwargs_map={'start_date': 'StartDate', 'end_date': 'EndDate', 'start_time': 'StartTime', 'end_time': 'EndTime'},
            func_params=['APIKey', 'start_date', 'end_date', 'start_time', 'end_time', 'ServiceType'],
            start_date=start_date,
            end_date=end_date,
        )
        
        return df
    
    
    def get_B1520(
        self,
        start_date: str='2020-01-01', 
        end_date: str='2020-01-07',
    ):
        """
        Changes In Actual Availability of Generation Units
        
        Parameters:
            start_date (str)
            end_date (str)
        """
        
        df = orchestrator.query_orchestrator(
            method='get_B1520',
            api_key=self.api_key,
            request_type='date_time_range',
            kwargs_map={'start_date': 'StartDate', 'end_date': 'EndDate', 'start_time': 'StartTime', 'end_time': 'EndTime'},
            func_params=['APIKey', 'start_date', 'end_date', 'start_time', 'end_time', 'ServiceType'],
            start_date=start_date,
            end_date=end_date,
        )
        
        return df
    
    
    def get_B1530(
        self,
        start_date: str='2020-01-01', 
        end_date: str='2020-01-07',
    ):
        """
        Planned Unavailability of Production Units
        
        Parameters:
            start_date (str)
            end_date (str)
        """
        
        df = orchestrator.query_orchestrator(
            method='get_B1530',
            api_key=self.api_key,
            request_type='date_time_range',
            kwargs_map={'start_date': 'StartDate', 'end_date': 'EndDate', 'start_time': 'StartTime', 'end_time': 'EndTime'},
            func_params=['APIKey', 'start_date', 'end_date', 'start_time', 'end_time', 'ServiceType'],
            start_date=start_date,
            end_date=end_date,
        )
        
        return df
    
    
    def get_B1540(
        self,
        start_date: str='2020-01-01', 
        end_date: str='2020-01-07',
    ):
        """
        Changes In Actual Availability of Production Units
        
        Parameters:
            start_date (str)
            end_date (str)
        """
        
        df = orchestrator.query_orchestrator(
            method='get_B1540',
            api_key=self.api_key,
            request_type='date_time_range',
            kwargs_map={'start_date': 'StartDate', 'end_date': 'EndDate', 'start_time': 'StartTime', 'end_time': 'EndTime'},
            func_params=['APIKey', 'start_date', 'end_date', 'start_time', 'end_time', 'ServiceType'],
            start_date=start_date,
            end_date=end_date,
        )
        
        return df
    
    
    def get_B1610(
        self,
        start_date: str='2020-01-01', 
        end_date: str='2020-01-01 1:30',
        NGCBMUnitID: str='*',
    ):
        """
        Actual Generation Output per Generation Unit
        
        Parameters:
            start_date (str)
            end_date (str)
            NGCBMUnitID (str)
        """
        
        df = orchestrator.query_orchestrator(
            method='get_B1610',
            api_key=self.api_key,
            request_type='SP_and_date',
            kwargs_map={'date': 'SettlementDate', 'SP': 'Period'},
            func_params=['APIKey', 'date', 'SP', 'NGCBMUnitID', 'ServiceType'],
            start_date=start_date,
            end_date=end_date,
            NGCBMUnitID=NGCBMUnitID,
        )
        
        return df
    
    
    def get_B1620(
        self,
        start_date: str='2020-01-01', 
        end_date: str='2020-01-01 1:30',
    ):
        """
        Actual Aggregated Generation per Type
        
        Parameters:
            start_date (str)
            end_date (str)
        """
        
        df = orchestrator.query_orchestrator(
            method='get_B1620',
            api_key=self.api_key,
            request_type='SP_and_date',
            kwargs_map={'date': 'SettlementDate', 'SP': 'Period'},
            func_params=['APIKey', 'date', 'SP', 'ServiceType'],
            start_date=start_date,
            end_date=end_date,
        )
        
        return df
    
    
    def get_B1630(
        self,
        start_date: str='2020-01-01', 
        end_date: str='2020-01-01 1:30',
    ):
        """
        Actual Or Estimated Wind and Solar Power Generation
        
        Parameters:
            start_date (str)
            end_date (str)
        """
        
        df = orchestrator.query_orchestrator(
            method='get_B1630',
            api_key=self.api_key,
            request_type='SP_and_date',
            kwargs_map={'date': 'SettlementDate', 'SP': 'Period'},
            func_params=['APIKey', 'date', 'SP', 'ServiceType'],
            start_date=start_date,
            end_date=end_date,
        )
        
        return df
    
    
    def get_B1720(
        self,
        start_date: str='2020-01-01', 
        end_date: str='2020-01-01 1:30',
    ):
        """
        Amount Of Balancing Reserves Under Contract Service
        
        Parameters:
            start_date (str)
            end_date (str)
        """
        
        df = orchestrator.query_orchestrator(
            method='get_B1720',
            api_key=self.api_key,
            request_type='SP_and_date',
            kwargs_map={'date': 'SettlementDate', 'SP': 'Period'},
            func_params=['APIKey', 'date', 'SP', 'ServiceType'],
            start_date=start_date,
            end_date=end_date,
        )
        
        return df
    
    
    def get_B1730(
        self,
        start_date: str='2020-01-01', 
        end_date: str='2020-01-01 1:30',
    ):
        """
        Prices Of Procured Balancing Reserves Service
        
        Parameters:
            start_date (str)
            end_date (str)
        """
        
        df = orchestrator.query_orchestrator(
            method='get_B1730',
            api_key=self.api_key,
            request_type='SP_and_date',
            kwargs_map={'date': 'SettlementDate', 'SP': 'Period'},
            func_params=['APIKey', 'date', 'SP', 'ServiceType'],
            start_date=start_date,
            end_date=end_date,
        )
        
        return df
    
    
    def get_B1740(
        self,
        start_date: str='2020-01-01', 
        end_date: str='2020-01-01 1:30',
    ):
        """
        Accepted Aggregated Offers
        
        Parameters:
            start_date (str)
            end_date (str)
        """
        
        df = orchestrator.query_orchestrator(
            method='get_B1740',
            api_key=self.api_key,
            request_type='SP_and_date',
            kwargs_map={'date': 'SettlementDate', 'SP': 'Period'},
            func_params=['APIKey', 'date', 'SP', 'ServiceType'],
            start_date=start_date,
            end_date=end_date,
        )
        
        return df
    
    
    def get_B1750(
        self,
        start_date: str='2020-01-01', 
        end_date: str='2020-01-01 1:30',
    ):
        """
        Activated Balancing Energy
        
        Parameters:
            start_date (str)
            end_date (str)
        """
        
        df = orchestrator.query_orchestrator(
            method='get_B1750',
            api_key=self.api_key,
            request_type='SP_and_date',
            kwargs_map={'date': 'SettlementDate', 'SP': 'Period'},
            func_params=['APIKey', 'date', 'SP', 'ServiceType'],
            start_date=start_date,
            end_date=end_date,
        )
        
        return df
    
    
    def get_B1760(
        self,
        start_date: str='2020-01-01', 
        end_date: str='2020-01-01 1:30',
    ):
        """
        Prices Of Activated Balancing Energy
        
        Parameters:
            start_date (str)
            end_date (str)
        """
        
        df = orchestrator.query_orchestrator(
            method='get_B1760',
            api_key=self.api_key,
            request_type='SP_and_date',
            kwargs_map={'date': 'SettlementDate', 'SP': 'Period'},
            func_params=['APIKey', 'date', 'SP', 'ServiceType'],
            start_date=start_date,
            end_date=end_date,
        )
        
        return df
    
    
    def get_B1770(
        self,
        start_date: str='2020-01-01', 
        end_date: str='2020-01-01 1:30',
    ):
        """
        Imbalance Prices
        
        Parameters:
            start_date (str)
            end_date (str)
        """
        
        df = orchestrator.query_orchestrator(
            method='get_B1770',
            api_key=self.api_key,
            request_type='SP_and_date',
            kwargs_map={'date': 'SettlementDate', 'SP': 'Period'},
            func_params=['APIKey', 'date', 'SP', 'ServiceType'],
            start_date=start_date,
            end_date=end_date,
        )
        
        return df
    
    
    def get_B1780(
        self,
        start_date: str='2020-01-01', 
        end_date: str='2020-01-01 1:30',
    ):
        """
        Aggregated Imbalance Volumes
        
        Parameters:
            start_date (str)
            end_date (str)
        """
        
        df = orchestrator.query_orchestrator(
            method='get_B1780',
            api_key=self.api_key,
            request_type='SP_and_date',
            kwargs_map={'date': 'SettlementDate', 'SP': 'Period'},
            func_params=['APIKey', 'date', 'SP', 'ServiceType'],
            start_date=start_date,
            end_date=end_date,
        )
        
        return df
    
    
    def get_B1790(
        self,
        start_date: str='2020-01-01', 
        end_date: str='2020-06-01',
    ):
        """
        Financial Expenses and Income For Balancing
        
        Parameters:
            start_date (str)
            end_date (str)
        """
        
        df = orchestrator.query_orchestrator(
            method='get_B1790',
            api_key=self.api_key,
            request_type='year_and_month',
            kwargs_map={'year': 'Year', 'month': 'Month'},
            func_params=['APIKey', 'year', 'month', 'ServiceType'],
            start_date=start_date,
            end_date=end_date,
        )
        
        return df
    
    
    def get_B1810(
        self,
        start_date: str='2020-01-01', 
        end_date: str='2020-01-01 1:30',
    ):
        """
        Cross-Border Balancing Volumes of Exchanged Bids and Offers
        
        Parameters:
            start_date (str)
            end_date (str)
        """
        
        df = orchestrator.query_orchestrator(
            method='get_B1810',
            api_key=self.api_key,
            request_type='SP_and_date',
            kwargs_map={'date': 'SettlementDate', 'SP': 'Period'},
            func_params=['APIKey', 'date', 'SP', 'ServiceType'],
            start_date=start_date,
            end_date=end_date,
        )
        
        return df
    
    
    def get_B1820(
        self,
        start_date: str='2020-01-01', 
        end_date: str='2020-01-01 1:30',
    ):
        """
        Cross-Border Balancing Prices
        
        Parameters:
            start_date (str)
            end_date (str)
        """
        
        df = orchestrator.query_orchestrator(
            method='get_B1820',
            api_key=self.api_key,
            request_type='SP_and_date',
            kwargs_map={'date': 'SettlementDate', 'SP': 'Period'},
            func_params=['APIKey', 'date', 'SP', 'ServiceType'],
            start_date=start_date,
            end_date=end_date,
        )
        
        return df
    
    
    def get_B1830(
        self,
        start_date: str='2020-01-01', 
        end_date: str='2020-01-01 1:30',
    ):
        """
        Cross-border Balancing Energy Activated
        
        Parameters:
            start_date (str)
            end_date (str)
        """
        
        df = orchestrator.query_orchestrator(
            method='get_B1830',
            api_key=self.api_key,
            request_type='SP_and_date',
            kwargs_map={'date': 'SettlementDate', 'SP': 'Period'},
            func_params=['APIKey', 'date', 'SP', 'ServiceType'],
            start_date=start_date,
            end_date=end_date,
        )
        
        return df
    
    
    def get_BOD(
        self,
        start_date: str='2020-01-01', 
        end_date: str='2020-01-01 1:30',
        BMUnitId: str='2__AEENG000, G, E.ON Energy, Solutions Limited, EAS-EST01',
        BMUnitType: str='G, S, E, I, T, etc',
        LeadPartyName: str='AES New Energy Limited',
        NGCBMUnit: str='EAS-ASP01, AES New Energy Limited, G, 2__AAEPD000',
        Name: str='2__AAEPD000',
    ):
        """
        Bid Offer Level Data
        
        Parameters:
            start_date (str)
            end_date (str)
            BMUnitId (str)
            BMUnitType (str)
            LeadPartyName (str)
            NGCBMUnit (str)
            Name (str)
        """
        
        df = orchestrator.query_orchestrator(
            method='get_BOD',
            api_key=self.api_key,
            request_type='SP_and_date',
            kwargs_map={'date': 'SettlementDate', 'SP': 'SettlementPeriod'},
            func_params=['APIKey', 'date', 'SP', 'BMUnitId', 'BMUnitType', 'LeadPartyName', 'NGCBMUnit', 'Name', 'ServiceType'],
            start_date=start_date,
            end_date=end_date,
            BMUnitId=BMUnitId,
            BMUnitType=BMUnitType,
            LeadPartyName=LeadPartyName,
            NGCBMUnit=NGCBMUnit,
            Name=Name,
        )
        
        return df
    
    
    def get_CDN(
        self,
        FromClearedDate: str='2021-01-01',
        ToClearedDate: str='2021-01-02',
    ):
        """
        Credit Default Notice Data
        
        Parameters:
            FromClearedDate (str)
            ToClearedDate (str)
        """
        
        df = orchestrator.query_orchestrator(
            method='get_CDN',
            api_key=self.api_key,
            request_type='non_temporal',
            kwargs_map={},
            func_params=['APIKey', 'FromClearedDate', 'ToClearedDate', 'ServiceType'],
            FromClearedDate=FromClearedDate,
            ToClearedDate=ToClearedDate,
        )
        
        return df
    
    
    def get_DETSYSPRICES(
        self,
        start_date: str='2020-01-01', 
        end_date: str='2020-01-01 1:30',
    ):
        """
        Detailed System Prices
        
        Parameters:
            start_date (str)
            end_date (str)
        """
        
        df = orchestrator.query_orchestrator(
            method='get_DETSYSPRICES',
            api_key=self.api_key,
            request_type='SP_and_date',
            kwargs_map={'date': 'SettlementDate', 'SP': 'SettlementPeriod'},
            func_params=['APIKey', 'date', 'SP', 'ServiceType'],
            start_date=start_date,
            end_date=end_date,
        )
        
        return df
    
    
    def get_DEVINDOD(
        self,
        start_date: str='2020-01-01', 
        end_date: str='2020-01-07',
    ):
        """
        Daily Energy Volume Data
        
        Parameters:
            start_date (str)
            end_date (str)
        """
        
        df = orchestrator.query_orchestrator(
            method='get_DEVINDOD',
            api_key=self.api_key,
            request_type='date_range',
            kwargs_map={'start_date': 'FromDate', 'end_date': 'ToDate'},
            func_params=['APIKey', 'start_date', 'end_date', 'ServiceType'],
            start_date=start_date,
            end_date=end_date,
        )
        
        return df
    
    
    def get_DISBSAD(
        self,
        start_date: str='2020-01-01', 
        end_date: str='2020-01-01 1:30',
    ):
        """
        Balancing Services Adjustment Action Data
        
        Parameters:
            start_date (str)
            end_date (str)
        """
        
        df = orchestrator.query_orchestrator(
            method='get_DISBSAD',
            api_key=self.api_key,
            request_type='SP_and_date',
            kwargs_map={'date': 'SettlementDate', 'SP': 'SettlementPeriod'},
            func_params=['APIKey', 'date', 'SP', 'ServiceType'],
            start_date=start_date,
            end_date=end_date,
        )
        
        return df
    
    
    def get_FORDAYDEM(
        self,
        start_date: str='2020-01-01', 
        end_date: str='2020-01-07',
        ZoneIdentifier: str='N',
    ):
        """
        Forecast Day and Day Ahead Demand Data
        
        Parameters:
            start_date (str)
            end_date (str)
            ZoneIdentifier (str)
        """
        
        df = orchestrator.query_orchestrator(
            method='get_FORDAYDEM',
            api_key=self.api_key,
            request_type='date_range',
            kwargs_map={'start_date': 'FromDate', 'end_date': 'ToDate'},
            func_params=['APIKey', 'ZoneIdentifier', 'start_date', 'end_date', 'ServiceType'],
            start_date=start_date,
            end_date=end_date,
            ZoneIdentifier=ZoneIdentifier,
        )
        
        return df
    
    
    def get_FREQ(
        self,
        FromDateTime: str='2021-01-01 00:01:00',
        ToDateTime: str='2021-02-01 23:59:00',
    ):
        """
        Rolling System Frequency
        
        Parameters:
            FromDateTime (str)
            ToDateTime (str)
        """
        
        df = orchestrator.query_orchestrator(
            method='get_FREQ',
            api_key=self.api_key,
            request_type='non_temporal',
            kwargs_map={},
            func_params=['APIKey', 'FromDateTime', 'ToDateTime', 'ServiceType'],
            FromDateTime=FromDateTime,
            ToDateTime=ToDateTime,
        )
        
        return df
    
    
    def get_FUELHH(
        self,
        start_date: str='2020-01-01', 
        end_date: str='2020-01-07',
    ):
        """
        Half Hourly Outturn Generation by Fuel Type
        
        Parameters:
            start_date (str)
            end_date (str)
        """
        
        df = orchestrator.query_orchestrator(
            method='get_FUELHH',
            api_key=self.api_key,
            request_type='date_range',
            kwargs_map={'start_date': 'FromDate', 'end_date': 'ToDate'},
            func_params=['APIKey', 'start_date', 'end_date', 'ServiceType'],
            start_date=start_date,
            end_date=end_date,
        )
        
        return df
    
    
    def get_MELIMBALNGC(
        self,
        start_date: str='2020-01-01', 
        end_date: str='2020-01-07',
        ZoneIdentifier: str='N',
    ):
        """
        Forecast Day and Day Ahead Margin and Imbalance Data
        
        Parameters:
            start_date (str)
            end_date (str)
            ZoneIdentifier (str)
        """
        
        df = orchestrator.query_orchestrator(
            method='get_MELIMBALNGC',
            api_key=self.api_key,
            request_type='date_range',
            kwargs_map={'start_date': 'FromDate', 'end_date': 'ToDate'},
            func_params=['APIKey', 'ZoneIdentifier', 'start_date', 'end_date', 'ServiceType'],
            start_date=start_date,
            end_date=end_date,
            ZoneIdentifier=ZoneIdentifier,
        )
        
        return df
    
    
    def get_MID(
        self,
        start_date: str='2020-01-01', 
        end_date: str='2020-01-07',
    ):
        """
        Market Index Data
        
        Parameters:
            start_date (str)
            end_date (str)
        """
        
        df = orchestrator.query_orchestrator(
            method='get_MID',
            api_key=self.api_key,
            request_type='date_range',
            kwargs_map={'start_date': 'FromSettlementDate', 'end_date': 'ToSettlementDate', 'SP': 'Period'},
            func_params=['APIKey', 'start_date', 'end_date', 'SP', 'ServiceType'],
            start_date=start_date,
            end_date=end_date,
        )
        
        return df
    
    
    def get_MessageDetailRetrieval(
        self,
        MessageId: str='',
        ParticipantId: str='',
        SequenceId: str='',
        ActiveFlag: str='N',
    ):
        """
        REMIT Flow - Message List Retrieval
        
        Parameters:
            MessageId (str)
            ParticipantId (str)
            SequenceId (str)
            ActiveFlag (str)
        """
        
        df = orchestrator.query_orchestrator(
            method='get_MessageDetailRetrieval',
            api_key=self.api_key,
            request_type='non_temporal',
            kwargs_map={},
            func_params=['APIKey', 'MessageId', 'ParticipantId', 'SequenceId', 'ActiveFlag', 'ServiceType'],
            MessageId=MessageId,
            ParticipantId=ParticipantId,
            SequenceId=SequenceId,
            ActiveFlag=ActiveFlag,
        )
        
        return df
    
    
    def get_MessageListRetrieval(
        self,
        EventStart: str='2021-01-01',
        EventEnd: str='2021-01-02',
        PublicationFrom: str='2021-01-01',
        PublicationTo: str='2021-01-02',
        ParticipantId: str='',
        MessageID: str='',
        AssetID: str='',
        EventType: str='',
        FuelType: str='',
        MessageType: str='',
        UnavailabilityType: str='',
    ):
        """
        REMIT Flow - Message List Retrieval
        
        Parameters:
            EventStart (str)
            EventEnd (str)
            PublicationFrom (str)
            PublicationTo (str)
            ParticipantId (str)
            MessageID (str)
            AssetID (str)
            EventType (str)
            FuelType (str)
            MessageType (str)
            UnavailabilityType (str)
        """
        
        df = orchestrator.query_orchestrator(
            method='get_MessageListRetrieval',
            api_key=self.api_key,
            request_type='non_temporal',
            kwargs_map={},
            func_params=['APIKey', 'EventStart', 'EventEnd', 'PublicationFrom', 'PublicationTo', 'ParticipantId', 'MessageID', 'AssetID', 'EventType', 'FuelType', 'MessageType', 'UnavailabilityType', 'ServiceType'],
            EventStart=EventStart,
            EventEnd=EventEnd,
            PublicationFrom=PublicationFrom,
            PublicationTo=PublicationTo,
            ParticipantId=ParticipantId,
            MessageID=MessageID,
            AssetID=AssetID,
            EventType=EventType,
            FuelType=FuelType,
            MessageType=MessageType,
            UnavailabilityType=UnavailabilityType,
        )
        
        return df
    
    
    def get_NETBSAD(
        self,
        start_date: str='2020-01-01', 
        end_date: str='2020-01-01 1:30',
        isTwoDayWindow: str='FALSE',
    ):
        """
        Balancing Service Adjustment Data
        
        Parameters:
            start_date (str)
            end_date (str)
            isTwoDayWindow (str)
        """
        
        df = orchestrator.query_orchestrator(
            method='get_NETBSAD',
            api_key=self.api_key,
            request_type='SP_and_date',
            kwargs_map={'date': 'SettlementDate', 'SP': 'SettlementPeriod'},
            func_params=['APIKey', 'date', 'SP', 'isTwoDayWindow', 'ServiceType'],
            start_date=start_date,
            end_date=end_date,
            isTwoDayWindow=isTwoDayWindow,
        )
        
        return df
    
    
    def get_NONBM(
        self,
        start_date: str='2020-01-01', 
        end_date: str='2020-01-07',
    ):
        """
        Non BM STOR Instructed Volume Data
        
        Parameters:
            start_date (str)
            end_date (str)
        """
        
        df = orchestrator.query_orchestrator(
            method='get_NONBM',
            api_key=self.api_key,
            request_type='date_range',
            kwargs_map={'start_date': 'FromDate', 'end_date': 'ToDate'},
            func_params=['start_date', 'end_date', 'APIKey', 'ServiceType'],
            start_date=start_date,
            end_date=end_date,
        )
        
        return df
    
    
    def get_PHYBMDATA(
        self,
        start_date: str='2020-01-01', 
        end_date: str='2020-01-01 1:30',
        BMUnitId: str='',
        BMUnitType: str='',
        LeadPartyName: str='',
        NGCBMUnit: str='',
        Name: str='',
    ):
        """
        Physical Data
        
        Parameters:
            start_date (str)
            end_date (str)
            BMUnitId (str)
            BMUnitType (str)
            LeadPartyName (str)
            NGCBMUnit (str)
            Name (str)
        """
        
        df = orchestrator.query_orchestrator(
            method='get_PHYBMDATA',
            api_key=self.api_key,
            request_type='SP_and_date',
            kwargs_map={'date': 'SettlementDate', 'SP': 'SettlementPeriod'},
            func_params=['APIKey', 'date', 'SP', 'BMUnitId', 'BMUnitType', 'LeadPartyName', 'NGCBMUnit', 'Name', 'ServiceType'],
            start_date=start_date,
            end_date=end_date,
            BMUnitId=BMUnitId,
            BMUnitType=BMUnitType,
            LeadPartyName=LeadPartyName,
            NGCBMUnit=NGCBMUnit,
            Name=Name,
        )
        
        return df
    
    
    def get_SYSDEM(
        self,
        start_date: str='2020-01-01', 
        end_date: str='2020-01-07',
    ):
        """
        System Demand
        
        Parameters:
            start_date (str)
            end_date (str)
        """
        
        df = orchestrator.query_orchestrator(
            method='get_SYSDEM',
            api_key=self.api_key,
            request_type='date_range',
            kwargs_map={'start_date': 'FromDate', 'end_date': 'ToDate'},
            func_params=['APIKey', 'start_date', 'end_date', 'ServiceType'],
            start_date=start_date,
            end_date=end_date,
        )
        
        return df
    
    
    def get_SYSWARN(
        self,
        start_date: str='2020-01-01', 
        end_date: str='2020-01-07',
    ):
        """
        System Warnings
        
        Parameters:
            start_date (str)
            end_date (str)
        """
        
        df = orchestrator.query_orchestrator(
            method='get_SYSWARN',
            api_key=self.api_key,
            request_type='date_range',
            kwargs_map={'start_date': 'FromDate', 'end_date': 'ToDate'},
            func_params=['APIKey', 'start_date', 'end_date', 'ServiceType'],
            start_date=start_date,
            end_date=end_date,
        )
        
        return df
    
    
    def get_TEMP(
        self,
        start_date: str='2020-01-01', 
        end_date: str='2020-01-07',
    ):
        """
        Temperature Data
        
        Parameters:
            start_date (str)
            end_date (str)
        """
        
        df = orchestrator.query_orchestrator(
            method='get_TEMP',
            api_key=self.api_key,
            request_type='date_range',
            kwargs_map={'start_date': 'FromDate', 'end_date': 'ToDate'},
            func_params=['APIKey', 'start_date', 'end_date', 'ServiceType'],
            start_date=start_date,
            end_date=end_date,
        )
        
        return df
    
    
    def get_WINDFORFUELHH(
        self,
        start_date: str='2020-01-01', 
        end_date: str='2020-01-07',
    ):
        """
        Wind Generation Forecast and Out-turn Data
        
        Parameters:
            start_date (str)
            end_date (str)
        """
        
        df = orchestrator.query_orchestrator(
            method='get_WINDFORFUELHH',
            api_key=self.api_key,
            request_type='date_range',
            kwargs_map={'start_date': 'FromDate', 'end_date': 'ToDate'},
            func_params=['APIKey', 'start_date', 'end_date', 'ServiceType'],
            start_date=start_date,
            end_date=end_date,
        )
        
        return df
    
    